/*
 * SSD1306/SSD1315に７セグメントっぽいフォントで表示
 * 
 */
#include <Wire.h>
#include "7SEG_DATA.h"
#include "SSD1306.h"

// SSD1306の初期化コマンド群
static uint8_t SSD1306_WIDTH = 128;  // OLEDの幅(ピクセル数)
static uint8_t SSD1306_HIGHT = 64;   // OLEDの高さ(ピクセル数)
static uint8_t SSD1306_PAGES = 8;    // OLEDのページ数(ピクセル数÷８)
static uint8_t SSD1306_ADDR = 0x3C;  // SSD1306のI2Cアドレス
static uint8_t SSD1306_CMD  = 0x80;  // SSD1306の1Byteコマンド
static uint8_t SSD1306_CMDS = 0x00;  // SSD1306の複数Byteコマンド

static uint8_t SSD1306_DATA = 0x40;  // SSD1306のデータバイト
static uint8_t SSD1306_INIT[] = {
  0xC0,         // C0/C8:上下の描画方向
  0xA0,         // A0/A1:左右の描画方向
  0x22,         // 20/21/22/23:アドレッシングモード
  0xA8,0x3F,    // 画面の解像度
  0xD3,0x00,    // 縦方向のオフセット
  0x40,         // 縦方向の描画開始位置
  0xA6,         // 表示方法 - A7にすると画面反転
  0x8D, 0x14    // OLEDパネルの供給電圧と思われる 14hに設定しないとダメらしい
};

// 初期化
void SSD1306::init() {
  // OLED設定コマンド送信
  oledSendCommands(sizeof(SSD1306_INIT), SSD1306_INIT) ;  // 初期化
  oledSendCommand(0xAF) ; // 画面表示
  cls() ;
}

void SSD1306::cls() {
  tftFillData(0) ;
  memset(displayBuff,0,sizeof(displayBuff)) ;
}

// TFTにコマンドを送信
void SSD1306::oledSendCommand(uint8_t command) {
  Wire.beginTransmission(SSD1306_ADDR);
  Wire.write(SSD1306_CMD);
  Wire.write(command);
  Wire.endTransmission();
}

// TFTにコマンド+1バイトデータを送信
void SSD1306::oledSendCommand(uint8_t command, uint8_t data1) {
  Wire.beginTransmission(SSD1306_ADDR);
  Wire.write(SSD1306_CMDS);
  Wire.write(command);
  Wire.write(data1);
  Wire.endTransmission();
}

// TFTにコマンド+2バイトデータを送信
void SSD1306::oledSendCommand(uint8_t command, uint8_t data1, uint8_t data2) {
  Wire.beginTransmission(SSD1306_ADDR);
  Wire.write(SSD1306_CMDS);
  Wire.write(command);
  Wire.write(data1);
  Wire.write(data2);
  Wire.endTransmission();
}

// TFTにコマンド+3バイトデータを送信
void SSD1306::oledSendCommand(uint8_t command, uint8_t data1, uint8_t data2, uint8_t data3) {
  Wire.beginTransmission(SSD1306_ADDR);
  Wire.write(SSD1306_CMDS);
  Wire.write(command);
  Wire.write(data1);
  Wire.write(data2);
  Wire.write(data3);
  Wire.endTransmission();
}

// TFTにコマンド+4バイトデータを送信
void SSD1306::oledSendCommand(uint8_t command, uint8_t data1, uint8_t data2, uint8_t data3, uint8_t data4) {
  Wire.beginTransmission(SSD1306_ADDR);
  Wire.write(SSD1306_CMDS);
  Wire.write(command);
  Wire.write(data1);
  Wire.write(data2);
  Wire.write(data4);
  Wire.endTransmission();
}

// OLEDに複数のコマンドを送信
void SSD1306::oledSendCommands(int commandLen,uint8_t commands[]) {
  Wire.beginTransmission(SSD1306_ADDR);
  Wire.write(SSD1306_CMDS);
  for (int i = 0; i < commandLen; i++) {
    Wire.write(commands[i]);
  }
   Wire.endTransmission();
}

// 画面を特定のビットバターンで埋める::画面の初期化用
void SSD1306::tftFillData(uint8_t fillData) {
  for (uint8_t m = 0; m < SSD1306_PAGES; m++) {
    oledSendCommand(0xB0 | (m & 0x07) , 0x10 , 0x00) ;

    Wire.beginTransmission(SSD1306_ADDR);
    Wire.write(SSD1306_DATA);
    for (uint8_t x = 0; x < SSD1306_WIDTH; x++) {
      if (Wire.write(fillData) == 0) {
        Wire.endTransmission();
        Wire.beginTransmission(SSD1306_ADDR);
        Wire.write(SSD1306_DATA);
        Wire.write(fillData);
      }
    }
    Wire.endTransmission();
  }
  oledSendCommand(0xB0 , 0x10 , 0x00) ;
}

// 7セグメント調フォント表示
void SSD1306::disp7SEG(int offsetX , int offsetY , int num) {
  if (num < 10) {
    bool flag7[7] ;
    for (int i=0;i<7;i++) {
      flag7[i] = addMap[num][i] ; 
    }
    for (int i = 0; i < 16*4; i++) {
      bitMapBuff[i] = 0 ;
      if (flag7[0]) { bitMapBuff[i] |= pgm_read_word(SEGa + i) ; }  // a
      if (flag7[1]) { bitMapBuff[i] |= pgm_read_word(SEGb + i) ; }  // b
      if (flag7[2]) { bitMapBuff[i] |= pgm_read_word(SEGc + i) ; }  // c
      if (flag7[3]) { bitMapBuff[i] |= pgm_read_word(SEGd + i) ; }  // d
      if (flag7[4]) { bitMapBuff[i] |= pgm_read_word(SEGe + i) ; }  // e
      if (flag7[5]) { bitMapBuff[i] |= pgm_read_word(SEGf + i) ; }  // f
      if (flag7[6]) { bitMapBuff[i] |= pgm_read_word(SEGg + i) ; }  // g
    }
  } else {
    uint8_t *bitMap ;
    switch(num) {
      case 10: bitMap = SEGdot ; break ; // DOT
      case 11: bitMap = SEGtemp ; break ; // 温度計
      case 12: bitMap = SEGoc ; break ; // ℃
      case 13: bitMap = SEGhum ; break ; // 湿度計
      case 14: bitMap = SEGp ; break ; // ％
      case 15: bitMap = SEGcoron ; break ; // ：

    }
    for (int i = 0; i < 16*4; i++) {
      bitMapBuff[i] = pgm_read_word(bitMap + i) ; 
    } 
  }

  oledBitMap(offsetX,offsetY, 16 , 4 , bitMapBuff) ;
}

// BITパターン表示
void SSD1306::oledBitMap(int offsetX , int offsetY , int width,int hight, uint8_t bitMap[]) {
  int pos = 0 ;
  for (int m = 0; m < hight; m++) {
    oledSendCommand(0xB0 | ((m + offsetY) & 0x07) , 0x10 | (offsetX & 0xf0) >> 4 , offsetX & 0x0F) ;
    Wire.beginTransmission(SSD1306_ADDR);
    Wire.write(SSD1306_DATA);
    for (int x = 0; x < width; x++) {
      int8_t byteData = bitMap[pos] ;
      if (Wire.write(byteData) == 0) {
        Wire.endTransmission();
        Wire.beginTransmission(SSD1306_ADDR);
        Wire.write(SSD1306_DATA);
        Wire.write(byteData);
      }
      pos++ ;
    }
    Wire.endTransmission();
  }
  oledSendCommand(0xB0 , 0x10 , 0x00) ;
}

void SSD1306::oledBitImage(int offsetX , int offsetY , int width,int hight, const uint8_t bitMap[]) {
  int pos = 0 ;
  for (int m = 0; m < hight; m++) {
    oledSendCommand(0xB0 | ((m + offsetY) & 0x07) , 0x10 | (offsetX & 0xf0) >> 4 , offsetX & 0x0F) ;

    Wire.beginTransmission(SSD1306_ADDR);
    Wire.write(SSD1306_DATA);
    for (int x = 0; x < width; x++) {
      int8_t byteData = pgm_read_word(bitMap + pos) ;
      if (Wire.write(byteData) == 0) {
        Wire.endTransmission();
        Wire.beginTransmission(SSD1306_ADDR);
        Wire.write(SSD1306_DATA);
        Wire.write(byteData);
      }
      pos++ ;
    }
    Wire.endTransmission();
  }
  oledSendCommand(0xB0 , 0x10 , 0x00) ;
}
void SSD1306::drawBuff( ) {
  int pos = 0 ;
  for (uint8_t m = 0; m < SSD1306_PAGES; m++) {
    oledSendCommand(0xB0 | (m & 0x07) , 0x10 , 0x00) ;

    Wire.beginTransmission(SSD1306_ADDR);
    Wire.write(SSD1306_DATA);
    for (uint8_t x = 0; x < SSD1306_WIDTH; x++) {
      uint8_t data = displayBuff[pos] ; pos ++ ;
      if (Wire.write(data) == 0) {
        Wire.endTransmission();
        Wire.beginTransmission(SSD1306_ADDR);
        Wire.write(SSD1306_DATA);
        Wire.write(data);
      }
    }
    Wire.endTransmission();
  }
  oledSendCommand(0xB0 , 0x10 , 0x00) ;
}

// --- Copy from Adafruit_SSD1306 Library
void SSD1306::drawPixel(int16_t x, int16_t y, uint8_t color) {
  if ((x >= 0) && (x < SSD1306_WIDTH) && (y >= 0) && (y < SSD1306_HIGHT)) {
    switch (color) {
    case SSD1306_WHITE:
      displayBuff[x + (y / 8) * SSD1306_WIDTH] |= (1 << (y & 7));
      break;
    case SSD1306_BLACK:
      displayBuff[x + (y / 8) * SSD1306_WIDTH] &= ~(1 << (y & 7));
      break;
    case SSD1306_INVERSE:
      displayBuff[x + (y / 8) * SSD1306_WIDTH] ^= (1 << (y & 7));
      break;
    }
  }
}

void SSD1306::drawFastHLine(int16_t x, int16_t y, int16_t w, uint8_t color) {
  if ((y >= 0) && (y < SSD1306_HIGHT)) { // Y coord in bounds?
    if (x < 0) {                  // Clip left
      w += x;
      x = 0;
    }
    if ((x + w) > SSD1306_WIDTH) { // Clip right
      w = (SSD1306_WIDTH - x);
    }
    if (w > 0) { // Proceed only if width is positive
      uint8_t *pBuf = &displayBuff[(y / 8) * SSD1306_WIDTH + x];
      uint8_t mask = 1 << (y & 7);
      switch (color) {
      case SSD1306_WHITE:
        while (w--) {
          *pBuf++ |= mask;
        };
        break;
      case SSD1306_BLACK:
        mask = ~mask;
        while (w--) {
          *pBuf++ &= mask;
        };
        break;
      case SSD1306_INVERSE:
        while (w--) {
          *pBuf++ ^= mask;
        };
        break;
      }
    }
  }
}

void SSD1306::drawFastVLine(int16_t x, int16_t __y, int16_t __h,uint8_t color) {
  if ((x >= 0) && (x < SSD1306_WIDTH)) { // X coord in bounds?
    if (__y < 0) {               // Clip top
      __h += __y;
      __y = 0;
    }
    if ((__y + __h) > SSD1306_HIGHT) { // Clip bottom
      __h = (SSD1306_HIGHT - __y);
    }
    if (__h > 0) { // Proceed only if height is now positive
      // this display doesn't need ints for coordinates,
      // use local byte registers for faster juggling
      int16_t y = __y, h = __h;
      uint8_t *pBuf = &displayBuff[(y / 8) * SSD1306_WIDTH + x];

      // do the first partial byte, if necessary - this requires some masking
      uint8_t mod = (y & 7);
      if (mod) {
        // mask off the high n bits we want to set
        mod = 8 - mod;
        // note - lookup table results in a nearly 10% performance
        // improvement in fill* functions
        // uint8_t mask = ~(0xFF >> mod);
        static uint8_t premask[8] = {0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE};
        uint8_t mask = premask[mod];
        // adjust the mask if we're not going to reach the end of this byte
        if (h < mod)
          mask &= (0XFF >> (mod - h));

        switch (color) {
        case SSD1306_WHITE:
          *pBuf |= mask;
          break;
        case SSD1306_BLACK:
          *pBuf &= ~mask;
          break;
        case SSD1306_INVERSE:
          *pBuf ^= mask;
          break;
        }
        pBuf += SSD1306_WIDTH;
      }

      if (h >= mod) { // More to go?
        h -= mod;
        // Write solid bytes while we can - effectively 8 rows at a time
        if (h >= 8) {
          if (color == SSD1306_INVERSE) {
            // separate copy of the code so we don't impact performance of
            // black/white write version with an extra comparison per loop
            do {
              *pBuf ^= 0xFF; // Invert byte
              pBuf += SSD1306_WIDTH; // Advance pointer 8 rows
              h -= 8;        // Subtract 8 rows from height
            } while (h >= 8);
          } else {
            // store a local value to work with
            uint8_t val = (color != SSD1306_BLACK) ? 255 : 0;
            do {
              *pBuf = val;   // Set byte
              pBuf += SSD1306_WIDTH; // Advance pointer 8 rows
              h -= 8;        // Subtract 8 rows from height
            } while (h >= 8);
          }
        }

        if (h) { // Do the final partial byte, if necessary
          mod = h & 7;
          // this time we want to mask the low bits of the byte,
          // vs the high bits we did above
          // uint8_t mask = (1 << mod) - 1;
          // note - lookup table results in a nearly 10% performance
          // improvement in fill* functions
          static uint8_t postmask[8] = {0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F};
          uint8_t mask = postmask[mod];
          switch (color) {
          case SSD1306_WHITE:
            *pBuf |= mask;
            break;
          case SSD1306_BLACK:
            *pBuf &= ~mask;
            break;
          case SSD1306_INVERSE:
            *pBuf ^= mask;
            break;
          }
        }
      }
    } // endif positive height
  }   // endif x in bounds
}

void SSD1306::drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1,uint8_t color) {
  // Update in subclasses if desired!
  if (x0 == x1) {
    if (y0 > y1)
      _swap_int8_t(y0, y1);
    drawFastVLine(x0, y0, y1 - y0 + 1, color);
  } else if (y0 == y1) {
    if (x0 > x1)
      _swap_int8_t(x0, x1);
    drawFastHLine(x0, y0, x1 - x0 + 1, color);
  } else {
    writeLine(x0, y0, x1, y1, color);
  }
  drawBuff( ) ;
}

void SSD1306::drawBox(int16_t x0, int16_t y0, int16_t x1, int16_t y1,uint8_t color) {
  // Update in subclasses if desired!
  if (y0 > y1) _swap_int8_t(y0, y1);
  if (x0 > x1) _swap_int8_t(x0, x1);
  drawFastVLine(x0, y0, y1 - y0 + 1, color);
  drawFastHLine(x0, y0, x1 - x0 + 1, color);
  drawFastVLine(x1, y0, y1 - y0 + 1, color);
  drawFastHLine(x0, y1, x1 - x0 + 1, color);

  drawBuff( ) ;
}

void SSD1306::writeLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint8_t color) {
  int16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    _swap_int8_t(x0, y0);
    _swap_int8_t(x1, y1);
  }

  if (x0 > x1) {
    _swap_int8_t(x0, x1);
    _swap_int8_t(y0, y1);
  }

  int16_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);

  int16_t err = dx / 2;
  int16_t ystep;

  if (y0 < y1) {
    ystep = 1;
  } else {
    ystep = -1;
  }

  for (; x0 <= x1; x0++) {
    if (steep) {
      drawPixel(y0, x0, color);
    } else {
      drawPixel(x0, y0, color);
    }
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;
    }
  }
}
